@@ -28,8 +28,6 @@
  * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  * SOFTWARE.
- *
- * $Id: srq_pingpong.c 7484 2006-05-24 21:12:21Z roland $
  */
 
 #if HAVE_CONFIG_H
@@ -152,7 +150,9 @@ static struct pingpong_dest *pp_client_exch_dest(const char *servername, int por
 	int sockfd = -1;
 	struct pingpong_dest *rem_dest = NULL;
 
-	asprintf(&service, "%d", port);
+	if (asprintf(&service, "%d", port) < 0)
+		return NULL;
+
 	n = getaddrinfo(servername, service, &hints, &res);
 
 	if (n < 0) {
@@ -231,7 +231,9 @@ static struct pingpong_dest *pp_server_exch_dest(struct pingpong_context *ctx,
 	int sockfd = -1, connfd;
 	struct pingpong_dest *rem_dest = NULL;
 
-	asprintf(&service, "%d", port);
+	if (asprintf(&service, "%d", port) < 0)
+		return NULL;
+
 	n = getaddrinfo(NULL, service, &hints, &res);
 
 	if (n < 0) {
@@ -360,7 +362,7 @@ static struct pingpong_context *pp_init_ctx(struct ibv_device *ib_dev, int size,
 
 	ctx->mr = ibv_reg_mr(ctx->pd, ctx->buf, size, IBV_ACCESS_LOCAL_WRITE);
 	if (!ctx->mr) {
-		fprintf(stderr, "Couldn't allocate MR\n");
+		fprintf(stderr, "Couldn't register MR\n");
 		return NULL;
 	}
 
@@ -426,6 +428,55 @@ static struct pingpong_context *pp_init_ctx(struct ibv_device *ib_dev, int size,
 	return ctx;
 }
 
+int pp_close_ctx(struct pingpong_context *ctx, int num_qp)
+{
+	int i;
+
+	for (i = 0; i < num_qp; ++i) {
+		if (ibv_destroy_qp(ctx->qp[i])) {
+			fprintf(stderr, "Couldn't destroy QP[%d]\n", i);
+			return 1;
+		}
+	}
+
+	if (ibv_destroy_srq(ctx->srq)) {
+		fprintf(stderr, "Couldn't destroy SRQ\n");
+		return 1;
+	}
+
+	if (ibv_destroy_cq(ctx->cq)) {
+		fprintf(stderr, "Couldn't destroy CQ\n");
+		return 1;
+	}
+
+	if (ibv_dereg_mr(ctx->mr)) {
+		fprintf(stderr, "Couldn't deregister MR\n");
+		return 1;
+	}
+
+	if (ibv_dealloc_pd(ctx->pd)) {
+		fprintf(stderr, "Couldn't deallocate PD\n");
+		return 1;
+	}
+
+	if (ctx->channel) {
+		if (ibv_destroy_comp_channel(ctx->channel)) {
+			fprintf(stderr, "Couldn't destroy completion channel\n");
+			return 1;
+		}
+	}
+
+	if (ibv_close_device(ctx->context)) {
+		fprintf(stderr, "Couldn't release context\n");
+		return 1;
+	}
+
+	free(ctx->buf);
+	free(ctx);
+
+	return 0;
+}
+
 static int pp_post_recv(struct pingpong_context *ctx, int n)
 {
 	struct ibv_sge list = {
@@ -519,6 +570,7 @@ int main(int argc, char *argv[])
 	int                      rcnt, scnt;
 	int			 num_wc;
 	int                      i;
+	int                      num_cq_events = 0;
 
 	srand48(getpid() * time(NULL));
 
@@ -573,6 +625,7 @@ int main(int argc, char *argv[])
 				usage(argv[0]);
 				return 1;
 			}
+			break;
 
 		case 'q':
 			num_qp = strtol(optarg, NULL, 0);
@@ -712,6 +765,8 @@ int main(int argc, char *argv[])
 				return 1;
 			}
 
+			++num_cq_events;
+
 			if (ev_cq != ctx->cq) {
 				fprintf(stderr, "CQ event for unknown CQ %p\n", ev_cq);
 				return 1;
@@ -803,5 +858,13 @@ int main(int argc, char *argv[])
 		       iters, usec / 1000000., usec / iters);
 	}
 
+	ibv_ack_cq_events(ctx->cq, num_cq_events);
+
+	if (pp_close_ctx(ctx, num_qp))
+		return 1;
+
+	ibv_free_device_list(dev_list);
+	free(rem_dest);
+
 	return 0;
 }
